import os, sys
from typing import List, Optional, Tuple, Union
from pandas.util import hash_pandas_object
import numpy as np
import SimpleITK as sitk
import pandas as pd
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torchio
import torchmetrics
from tqdm import tqdm
from torchinfo import summary
from GANDLF.utils.generic import get_array_from_image_or_tensor

# global definition for both one_hot and reverse_one_hot
special_cases_to_check = ["||"]


def one_hot(
    segmask_tensor: torch.Tensor, class_list: Union[List[int], List[str]]
) -> torch.Tensor:
    """
    This function creates a one-hot-encoded mask from the segmentation mask Tensor and specified class list

    Args:
        segmask_tensor (torch.Tensor): The segmentation mask Tensor.
        class_list (Union[List[int], List[str]]): The list of classes based on which one-hot encoding needs to happen.

    Returns:
        torch.Tensor: The one-hot encoded torch.Tensor
    """
    batch_size = segmask_tensor.shape[0]

    def_shape = segmask_tensor.shape
    if len(def_shape) == 4:
        # special case for sdnet
        batch_stack = torch.zeros(
            def_shape[0],
            len(class_list),
            def_shape[2],
            def_shape[3],
            dtype=torch.float32,
            device=segmask_tensor.device,
        )
    else:
        batch_stack = torch.zeros(
            def_shape[0],
            len(class_list),
            def_shape[2],
            def_shape[3],
            def_shape[4],
            dtype=torch.float32,
            device=segmask_tensor.device,
        )

    for b in range(batch_size):
        # since the input tensor is 5D, with [batch_size, modality, x, y, z], we do not need to consider the modality dimension for labels
        segmask_array_iter = segmask_tensor[b, 0, ...]
        bin_mask = segmask_array_iter == 0  # initialize bin_mask

        # this implementation allows users to combine logical operands
        class_idx = 0
        for _class in class_list:
            if isinstance(_class, str):
                for case in special_cases_to_check:
                    if case in _class:
                        special_class_split = _class.split(case)
                        bin_mask = segmask_array_iter == int(special_class_split[0])
                        for i in range(1, len(special_class_split)):
                            bin_mask = torch.logical_or(
                                bin_mask,
                                (segmask_array_iter == int(special_class_split[i])),
                            )
                    else:
                        # assume that it is a simple int
                        bin_mask = segmask_array_iter == int(_class)
            else:
                bin_mask = segmask_array_iter == int(_class)
            # we always ensure the append happens in dim 0
            batch_stack[b, class_idx, ...] = bin_mask.long().unsqueeze(0)
            class_idx += 1

    return batch_stack


def reverse_one_hot(predmask_tensor: torch.Tensor, class_list: Union[List[int], List[str]]) -> np.array:
    """
    This function creates a full segmentation mask Tensor from a one-hot-encoded mask and specified class list

    Args:
        predmask_tensor (torch.Tensor): The predicted segmentation mask Tensor.
        class_list (Union[List[int], List[str]]): The list of classes based on which one-hot encoding needs to happen.

    Returns:
        numpy.array: The final mask as numpy array.
    """
    predmask_array = get_array_from_image_or_tensor(predmask_tensor)
    special_case_detected = False

    for _class in class_list:
        for case in special_cases_to_check:
            if isinstance(_class, str):
                if case in _class:  # check if any of the special cases are present
                    special_case_detected = True
                    break

    final_mask = np.zeros(predmask_array[0, ...].shape).astype(np.int16)
    predmask_array_bool = predmask_array >= 0.5

    # for special case, do not use '0' to initialize any value in final_mask in case it is absent
    zero_present = False
    if special_case_detected:
        for _class in class_list:
            if (_class == "0") or (_class == 0):
                zero_present = True
                break
    for idx, i in enumerate(class_list):
        output_value = i
        # for special case, use the index as value
        if special_case_detected:
            output_value = idx
            # if zero is not present, then don't use '0' as output value
            if not (zero_present):
                output_value += 1

        final_mask[predmask_array_bool[idx, ...]] = output_value

    return final_mask


def send_model_to_device(
    model: torch.nn.Module, amp: bool, device: str, optimizer: torch.optim
) -> Union[torch.nn.Module, bool, torch.device, int]:
    """
    This function reads the environment variable(s) and send model to correct device

    Args:
        model (torch.nn.Module): The model that needs to be sent to specified device.
        amp (bool): Whether automatic mixed precision is to be used.
        device (str): Device type.
        optimizer (torch.optim): The optimizer for training.

    Returns:
        torch.nn.Module: The model after it has been sent to specified device
        bool: Whether automatic mixed precision is to be used or not.
        torch.device: Device type.
    """
    if device == "cuda":
        assert torch.cuda.is_available(), "CUDA is either not available or not enabled"
        assert (
            os.environ.get("CUDA_VISIBLE_DEVICES") is not None
        ), "CUDA_VISIBLE_DEVICES is not set"

        dev = os.environ.get("CUDA_VISIBLE_DEVICES")
        # multi-gpu support
        # ###
        # # https://discuss.pytorch.org/t/cuda-visible-devices-make-gpu-disappear/21439/17?u=sarthakpati
        # ###
        if "," in dev:
            device = torch.device("cuda")
            dev_to_pass_to_torch = [*range(len(dev.split(",")))]
            model = nn.DataParallel(model, device_ids=dev_to_pass_to_torch)
            ## this is the new api, but it is a bit finicky and needs further testing
            # model = nn.parallel.DistributedDataParallel(
            #     model,
            #     device_ids=dev_to_pass_to_torch,
            #     output_device=dev_to_pass_to_torch[0],
            # )
        else:
            print("Device requested via CUDA_VISIBLE_DEVICES: ", dev)
            print("Total number of CUDA devices: ", torch.cuda.device_count())

            # if only a single visible device, it will be indexed as '0'
            if torch.cuda.device_count() == 1:
                dev = "0"

            dev_int = int(dev)
            print("Device finally used: ", dev)
            # device = torch.device('cuda:' + dev)
            device = torch.device("cuda")
            print("Sending model to aforementioned device")
            model = model.to(device)
            print(
                "Memory Total : ",
                round(
                    torch.cuda.get_device_properties(dev_int).total_memory / 1024**3,
                    1,
                ),
                "GB, Allocated: ",
                round(torch.cuda.memory_allocated(dev_int) / 1024**3, 1),
                "GB, Cached: ",
                round(torch.cuda.memory_reserved(dev_int) / 1024**3, 1),
                "GB",
            )

        print(
            "Device - Current: %s Count: %d Name: %s Availability: %s"
            % (
                torch.cuda.current_device(),
                torch.cuda.device_count(),
                torch.cuda.get_device_name(device),
                torch.cuda.is_available(),
            )
        )

        if not (optimizer is None):
            # ensuring optimizer is in correct device - https://github.com/pytorch/pytorch/issues/8741
            optimizer.load_state_dict(optimizer.state_dict())

    else:
        dev = -1
        device = torch.device("cpu")
        model.cpu()
        amp = False
        print("Since Device is CPU, Mixed Precision Training is set to False")

    return model, amp, device, dev


def get_model_dict(model: torch.nn.Module, device_id: Union[str, List[str]]) -> dict:
    """
    This function returns the model dictionary

    Args:
        model (torch.nn.Module): The model for which the dictionary is to be returned.
        device_id (Union[str, List[str]]): The device id as string or list of devices.

    Returns:
        dict: The model dictionary.
    """
    multi_gpu_flag = True if isinstance(device_id, list) else False
    if isinstance(device_id, str):
        if "," in device_id:
            multi_gpu_flag = True
    if multi_gpu_flag:
        model_dict = model.module.state_dict()
    else:
        model_dict = model.state_dict()

    return model_dict


def get_class_imbalance_weights_classification(
    training_df: pd.DataFrame, params: dict
) -> Tuple[dict, dict, dict]:
    """
    This function calculates the penalty used for loss functions in multi-class problems.
    It looks at the column "valuesToPredict" and identifies unique classes, fetches the class distribution
    and generates the required class weights, and then generates the weights needed for loss function which
    are inverse of the class weights generated divided by the total number of items in a single column

    Args:
        training_Df (pd.DataFrame): The training data frame.
        parameters (dict) : The parameters passed by the user yaml.

    Returns:
        Tuple[dict, dict, dict]: The penalty weights, sampling weights, and class weights for different classes under consideration.
    """
    predictions_array = (
        training_df[training_df.columns[params["headers"]["predictionHeaders"]]]
        .to_numpy()
        .ravel()
    )
    class_count = np.bincount(predictions_array)
    classes_to_predict = np.unique(predictions_array)
    total_count = len(training_df)
    penalty_dict, weight_dict = {}, {}

    # for the classes that are present in the training set, construct the weights as needed
    for i in classes_to_predict:
        weight_dict[i] = (class_count[i] + sys.float_info.epsilon) / total_count
        penalty_dict[i] = (1 + sys.float_info.epsilon) / weight_dict[i]

    # this is a corner case
    # for the classes that are requested for training but aren't present in the training set, assign largest possible penalty
    for i in params["model"]["class_list"]:
        i = int(i)
        if i not in weight_dict:
            print(
                "WARNING: A class was found in 'class_list' that was not present in the training data, please re-check training data labels"
            )
            weight_dict[i] = sys.float_info.epsilon
            penalty_dict[i] = (1 + sys.float_info.epsilon) / weight_dict[i]

    # ensure sum of penalties is always 1
    penalty_sum = (
        np.fromiter(penalty_dict.values(), dtype=np.float64).sum()
        + sys.float_info.epsilon
    )
    for i in range(params["model"]["num_classes"]):
        penalty_dict[i] /= penalty_sum

    # passing None for sampling_weights because there is no clear way to calculate this for classification tasks which do not have a label
    return penalty_dict, None, weight_dict


def get_class_imbalance_weights_segmentation(
    training_data_loader: DataLoader, parameters: dict
) -> Tuple[dict, dict, dict]:
    """
    This function calculates the penalty that is used for validation loss in multi-class problems

    Args:
        training_data_loader (torch.utils.data.DataLoader): The training data loader.
        parameters (dict): The parameters passed by the user yaml.

    Returns:
        Tuple[dict, dict, dict]: The penalty weights, sampling weights, and class weights for different classes under consideration.
    """
    abs_dict = {}  # absolute counts for each class
    weights_dict = {}  # average for "weighted averaging"
    penalty_dict = None  # penalty for misclassification
    # basically, do this for segmentation/classification tasks

    if parameters["problem_type"] != "regression":
        penalty_dict = {}
        for i in range(0, len(parameters["model"]["class_list"])):
            abs_dict[i] = 0
            penalty_dict[i] = 0

    penalty_loader = training_data_loader

    # get the weights for use for dice loss
    total_counter = 0

    # For regression dice penalty need not be taken account
    # For classification this should be calculated on the basis of predicted labels and mask
    # iterate through full penalty data
    for _, (subject) in enumerate(
        tqdm(penalty_loader, desc="Looping over training data for penalty calculation")
    ):
        # accumulate dice weights for each label
        mask = subject["label"][torchio.DATA]
        one_hot_mask = one_hot(mask, parameters["model"]["class_list"])
        for i in range(0, len(parameters["model"]["class_list"])):
            currentNumber = torch.nonzero(one_hot_mask[:, i, ...], as_tuple=False).size(
                0
            )
            # class-specific non-zero voxels
            abs_dict[i] += currentNumber
            # total number of non-zero voxels to be considered
            total_counter += currentNumber

    # Normalize class weights
    weights_dict = {
        key: (val + sys.float_info.epsilon) / total_counter
        for key, val in abs_dict.items()
    }

    # get the raw penalty values
    penalty = {
        key: total_counter / (len(abs_dict) * (val + sys.float_info.epsilon))
        for key, val in abs_dict.items()
    }
    # normalize penalty to sum of 1
    penalty_sum = np.fromiter(penalty.values(), dtype=np.float64).sum()
    penalty_dict = {
        key: (val + sys.float_info.epsilon) / penalty_sum
        for key, val in penalty.items()
    }

    return penalty_dict, penalty_dict, weights_dict


def get_class_imbalance_weights(
    training_df: pd.DataFrame, params: dict
) -> Tuple[dict, dict, dict]:
    """
    This function calculates the penalty that is used for validation loss in multi-class problems

    Args:
        training_df (pd.DataFrame): The training data frame.
        params (dict): The parameters passed by the user yaml.

    Returns:
        Tuple[dict, dict, dict]: The penalty weights, sampling weights, and class weights for different classes under consideration.
    """
    penalty_weights, sampling_weights, class_weights = None, None, None
    if params["weighted_loss"] or params["patch_sampler"]["biased_sampling"]:
        (penalty_weights, sampling_weights, class_weights) = (
            params.get("penalty_weights", None),
            params.get("sampling_weights", None),
            params.get("class_weights", None),
        )
        # this default is needed for openfl
        params["previous_parameters"] = params.get("previous_parameters", None)
        if params["previous_parameters"] is not None:
            previous_training_hash = params["previous_parameters"]["training_data_hash"]
            current_training_data_hash = params.get(
                "training_data_hash", hash_pandas_object(training_df).sum()
            )
            # compare the previous and current training data hashes, and reset the weights if the training data has changed
            penalty_weights = (
                None
                if previous_training_hash != current_training_data_hash
                else penalty_weights
            )

        # calculate the penalty/sampling weights only if one of the following conditions are met
        if (params["weighted_loss"] and (penalty_weights is None)) or (
            params["patch_sampler"]["biased_sampling"] and (sampling_weights is None)
        ):
            print("Calculating weights")
            if params["problem_type"] == "classification":
                (
                    penalty_weights,
                    sampling_weights,
                    class_weights,
                ) = get_class_imbalance_weights_classification(training_df, params)
            elif params["problem_type"] == "segmentation":
                # Set up the dataloader for penalty calculation
                from GANDLF.data.ImagesFromDataFrame import ImagesFromDataFrame

                penalty_data = ImagesFromDataFrame(
                    training_df,
                    parameters=params,
                    train=False,
                    loader_type="penalty",
                )

                penalty_loader = DataLoader(
                    penalty_data,
                    batch_size=1,
                    shuffle=True,
                    pin_memory=False,
                )

                (
                    penalty_weights,
                    sampling_weights,
                    class_weights,
                ) = get_class_imbalance_weights_segmentation(penalty_loader, params)
                del penalty_data, penalty_loader
        else:
            print("Using weights from config file")

    return penalty_weights, sampling_weights, class_weights


def get_linear_interpolation_mode(dimensionality: int) -> str:
    """
    Get linear interpolation mode.

    Args:
        dimensionality (int): The dimensions based on which interpolation mode is calculated

    Returns:
        str: Interpolation type to pass to interpolation function
    """

    mode = "nearest"
    if dimensionality == 2:
        mode = "bilinear"
    elif dimensionality == 3:
        mode = "trilinear"

    return mode


def print_model_summary(
    model: torch.nn.Module,
    input_batch_size: int,
    input_num_channels: int,
    input_patch_size: tuple,
    device: Optional[torch.device] = None,
) -> None:
    """
    This function prints the model summary.

    Args:
        model (torch.nn.Module): The model for which the summary is to be printed.
        input_batch_size (int): The input batch size.
        input_num_channels (int): The input number of channels.
        input_patch_size (tuple): The input patch size.
        device (Optional[torch.device], optional): The device. Defaults to None.
    """
    input_size = (input_batch_size, input_num_channels) + tuple(input_patch_size)
    if input_size[-1] == 1:
        input_size = input_size[:-1]
    try:
        stats = summary(model, input_size, device=device, verbose=0)

        print("Model Summary:")
        print("\tInput size:", stats.to_megabytes(stats.total_input), "MB")
        print("\tOutput size:", stats.to_megabytes(stats.total_output_bytes), "MB")
        print("\tParameters size:", stats.to_megabytes(stats.total_param_bytes), "MB")
        print(
            "\tEstimated total size:",
            stats.to_megabytes(
                stats.total_input + stats.total_output_bytes + stats.total_param_bytes
            ),
            "MB",
        )
        temp_output = stats.to_readable(stats.total_mult_adds)
        print("\tTotal # of operations:", temp_output[1], temp_output[0])
    except Exception as e:
        print("Failed to generate model summary with error: ", e)


def get_ground_truths_and_predictions_tensor(
    params: dict, loader_type: str
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    This function is used to get the ground truths and predictions for a given loader type.

    Args:
        params (dict): The parameters passed by the user yaml.
        loader_type (str): The loader type for which the ground truths and predictions are to be returned.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: The ground truths and predictions for the given loader type.
    """
    ground_truth_array = torch.from_numpy(
        params[loader_type][
            params[loader_type].columns[params["headers"]["predictionHeaders"]]
        ]
        .to_numpy()
        .ravel()
    ).type(torch.int)
    predictions_array = torch.zeros_like(ground_truth_array)

    return ground_truth_array, predictions_array


def get_output_from_calculator(
    prediction: torch.Tensor, target: torch.tensor, calculator: torchmetrics.Metric
) -> float:
    """
    Helper function to get the output from a calculator.

    Args:
        predictions (torch.Tensor): The output of the model.
        ground_truth (torch.Tensor): The ground truth labels.
        calculator (torchmetrics.Metric): The calculator to use.

    Returns:
        float: The output from the calculator.
    """
    temp_output = calculator(prediction, target)
    if temp_output.dim() > 0:
        temp_output = temp_output.cpu().tolist()
    else:
        temp_output = temp_output.cpu().item()
    return temp_output


def get_tensor_from_image(input_image: Union[sitk.Image, str]) -> torch.Tensor:
    """
    This function converts a sitk image to a torch tensor.

    Args:
        input_image (sitk.Image): The input image.

    Returns:
        torch.Tensor: The converted torch tensor.
    """
    input_image = (
        sitk.ReadImage(input_image) if isinstance(input_image, str) else input_image
    )
    return torch.from_numpy(sitk.GetArrayFromImage(input_image))


def get_image_from_tensor(input_tensor: torch.Tensor) -> sitk.Image:
    """
    This function converts a torch tensor to a sitk image.

    Args:
        input_tensor (torch.Tensor): The input tensor.

    Returns:
        sitk.Image: The converted sitk image.
    """
    arr = input_tensor.cpu().numpy()
    return_image = sitk.GetImageFromArray(arr)
    # this is specifically the case for 3D images
    if (arr.shape[0] == 1) and (arr.shape[1] > 3):
        return_image = sitk.GetImageFromArray(arr[0])

    return return_image



import os, time, psutil
from typing import Tuple
import pandas as pd
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm
import numpy as np
import torchio
from medcam import medcam

from GANDLF.data import get_testing_loader
from GANDLF.grad_clipping.grad_scaler import GradScaler, model_parameters_exclude_head
from GANDLF.grad_clipping.clip_gradients import dispatch_clip_grad_
from GANDLF.utils import (
    get_date_time,
    best_model_path_end,
    latest_model_path_end,
    initial_model_path_end,
    save_model,
    optimize_and_save_model,
    load_model,
    version_check,
    write_training_patches,
    print_model_summary,
    get_ground_truths_and_predictions_tensor,
    get_model_dict,
    print_and_format_metrics,
)
from GANDLF.metrics import overall_stats
from GANDLF.logger import Logger
from .step import step
from .forward_pass import validate_network
from .generic import create_pytorch_objects

# hides torchio citation request, see https://github.com/fepegar/torchio/issues/235
os.environ["TORCHIO_HIDE_CITATION_PROMPT"] = "1"


def train_network(
    model: torch.nn.Module,
    train_dataloader: DataLoader,
    optimizer: torch.optim.Optimizer,
    params: dict,
) -> Tuple[float, dict]:
    """
    This function performs the training of the network.

    Args:
        model (torch.nn.Module): The model to process the input image with, it should support appropriate dimensions.
        train_dataloader (DataLoader): The dataloader for the training epoch.
        optimizer (torch.optim.Optimizer): Optimizer for optimizing network.
        params (dict): The parameters dictionary.

    Returns:
        Tuple[float, dict]: The average epoch training loss and metrics.
    """
    print("*" * 20)
    print("Starting Training : ")
    print("*" * 20)
    # Initialize a few things
    total_epoch_train_loss = 0
    total_epoch_train_metric = {}
    average_epoch_train_metric = {}
    calculate_overall_metrics = (params["problem_type"] == "classification") or (
        params["problem_type"] == "regression"
    )

    for metric in params["metrics"]:
        if "per_label" in metric:
            total_epoch_train_metric[metric] = []
        else:
            total_epoch_train_metric[metric] = 0

    # automatic mixed precision - https://pytorch.org/docs/stable/amp.html
    if params["model"]["amp"]:
        scaler = GradScaler()
        if params["verbose"]:
            print("Using Automatic mixed precision", flush=True)

    # get ground truths
    if calculate_overall_metrics:
        (
            ground_truth_array,
            predictions_array,
        ) = get_ground_truths_and_predictions_tensor(params, "training_data")
    # Set the model to train
    model.train()
    for batch_idx, (subject) in enumerate(
        tqdm(train_dataloader, desc="Looping over training data")
    ):
        optimizer.zero_grad()
        image = (
            torch.cat(
                [subject[key][torchio.DATA] for key in params["channel_keys"]], dim=1
            )
            .float()
            .to(params["device"])
        )
        if "value_keys" in params:
            label = torch.cat([subject[key] for key in params["value_keys"]], dim=0)
            # min is needed because for certain cases, batch size becomes smaller than the total remaining labels
            label = label.reshape(
                min(params["batch_size"], len(label)),
                len(params["value_keys"]),
            )
        else:
            label = subject["label"][torchio.DATA]
        label = label.to(params["device"])

        if params["save_training"]:
            write_training_patches(
                subject,
                params,
            )

        # ensure spacing is always present in params and is always subject-specific
        if "spacing" in subject:
            params["subject_spacing"] = subject["spacing"]
        else:
            params["subject_spacing"] = None
        loss, calculated_metrics, output, _ = step(model, image, label, params)
        # store predictions for classification
        if calculate_overall_metrics:
            predictions_array[
                batch_idx
                * params["batch_size"] : (batch_idx + 1)
                * params["batch_size"]
            ] = (torch.argmax(output[0], 0).cpu().item())

        nan_loss = torch.isnan(loss)
        second_order = (
            hasattr(optimizer, "is_second_order") and optimizer.is_second_order
        )
        if params["model"]["amp"]:
            with torch.cuda.amp.autocast():
                # if loss is nan, don't backprop and don't step optimizer
                if not nan_loss:
                    scaler(
                        loss=loss,
                        optimizer=optimizer,
                        clip_grad=params["clip_grad"],
                        clip_mode=params["clip_mode"],
                        parameters=model_parameters_exclude_head(
                            model, clip_mode=params["clip_mode"]
                        ),
                        create_graph=second_order,
                    )
        else:
            if not nan_loss:
                loss.backward(create_graph=second_order)
                if params["clip_grad"] is not None:
                    dispatch_clip_grad_(
                        parameters=model_parameters_exclude_head(
                            model, clip_mode=params["clip_mode"]
                        ),
                        value=params["clip_grad"],
                        mode=params["clip_mode"],
                    )
                optimizer.step()

        # Non network training related
        if not nan_loss:
            total_epoch_train_loss += loss.detach().cpu().item()
        for metric in calculated_metrics.keys():
            if isinstance(total_epoch_train_metric[metric], list):
                if len(total_epoch_train_metric[metric]) == 0:
                    total_epoch_train_metric[metric] = np.array(
                        calculated_metrics[metric]
                    )
                else:
                    total_epoch_train_metric[metric] += np.array(
                        calculated_metrics[metric]
                    )
            else:
                total_epoch_train_metric[metric] += calculated_metrics[metric]

        if params["verbose"]:
            # For printing information at halftime during an epoch
            if ((batch_idx + 1) % (len(train_dataloader) / 2) == 0) and (
                (batch_idx + 1) < len(train_dataloader)
            ):
                print(
                    "\nHalf-Epoch Average train loss : ",
                    total_epoch_train_loss / (batch_idx + 1),
                )
                for metric in params["metrics"]:
                    if isinstance(total_epoch_train_metric[metric], np.ndarray):
                        to_print = (
                            total_epoch_train_metric[metric] / (batch_idx + 1)
                        ).tolist()
                    else:
                        to_print = total_epoch_train_metric[metric] / (batch_idx + 1)
                    print(
                        "Half-Epoch Average train " + metric + " : ",
                        to_print,
                    )

    average_epoch_train_loss = total_epoch_train_loss / len(train_dataloader)
    print("     Epoch Final   train loss : ", average_epoch_train_loss)

    # get overall stats for classification
    if calculate_overall_metrics:
        average_epoch_train_metric = overall_stats(
            predictions_array, ground_truth_array, params
        )
    average_epoch_train_metric = print_and_format_metrics(
        average_epoch_train_metric,
        total_epoch_train_metric,
        params["metrics"],
        "train",
        len(train_dataloader),
    )

    return average_epoch_train_loss, average_epoch_train_metric


def training_loop(
    training_data: pd.DataFrame,
    validation_data: pd.DataFrame,
    device: str,
    params: dict,
    output_dir: str,
    testing_data: bool = None,
    epochs: bool = None,
) -> None:
    """
    The main training loop.

    Args:
        training_data (pd.DataFrame): The data to use for training.
        validation_data (pd.DataFrame): The data to use for validation.
        device (str): The device to perform computations on.
        params (dict): The parameters dictionary.
        output_dir (str): The output directory.
        testing_data (pd.DataFrame): The data to use for testing.
        epochs (int): The number of epochs to train; if None, take from params.
    """
    # Some autodetermined factors
    if epochs is None:
        epochs = params["num_epochs"]
    params["device"] = device
    params["output_dir"] = output_dir
    params["training_data"] = training_data
    params["validation_data"] = validation_data
    params["testing_data"] = testing_data
    testingDataDefined = True
    if params["testing_data"] is None:
        # testing_data = validation_data
        testingDataDefined = False

    # Setup a few variables for tracking
    best_loss = 1e7
    patience, start_epoch = 0, 0
    first_model_saved = False
    model_paths = {
        "best": os.path.join(
            output_dir, params["model"]["architecture"] + best_model_path_end
        ),
        "initial": os.path.join(
            output_dir, params["model"]["architecture"] + initial_model_path_end
        ),
        "latest": os.path.join(
            output_dir, params["model"]["architecture"] + latest_model_path_end
        ),
    }

    # if previous model file is present, load it up for sanity checks
    main_dict = None
    if os.path.exists(model_paths["best"]):
        main_dict = load_model(model_paths["best"], params["device"])
        version_check(params["version"], version_to_check=main_dict["version"])
        params["previous_parameters"] = main_dict.get("parameters", None)

    # Defining our model here according to parameters mentioned in the configuration file
    print("Number of channels : ", params["model"]["num_channels"])

    (
        model,
        optimizer,
        train_dataloader,
        val_dataloader,
        scheduler,
        params,
    ) = create_pytorch_objects(params, training_data, validation_data, device)

    # save the initial model
    if not os.path.exists(model_paths["initial"]):
        save_model(
            {
                "epoch": 0,
                "model_state_dict": get_model_dict(model, params["device_id"]),
                "optimizer_state_dict": optimizer.state_dict(),
                "loss": best_loss,
            },
            model,
            params,
            model_paths["initial"],
            onnx_export=False,
        )
        print("Initial model saved.")

    # if previous model file is present, load it up
    if main_dict is not None:
        try:
            model.load_state_dict(main_dict["model_state_dict"])
            start_epoch = main_dict["epoch"]
            optimizer.load_state_dict(main_dict["optimizer_state_dict"])
            best_loss = main_dict["loss"]
            params["previous_parameters"] = main_dict.get("parameters", None)
            print("Previous model successfully loaded.")
        except RuntimeWarning:
            RuntimeWarning("Previous model could not be loaded, initializing model")

    if params["model"]["print_summary"]:
        print_model_summary(
            model,
            params["batch_size"],
            params["model"]["num_channels"],
            params["patch_size"],
            params["device"],
        )

    if testingDataDefined:
        test_dataloader = get_testing_loader(params)

    # Start training time here
    start_time = time.time()

    if not (os.environ.get("HOSTNAME") is None):
        print("Hostname :", os.environ.get("HOSTNAME"))

    # datetime object containing current date and time
    print("Initializing training at :", get_date_time(), flush=True)

    calculate_overall_metrics = (params["problem_type"] == "classification") or (
        params["problem_type"] == "regression"
    )

    # get the overall metrics that are calculated automatically for classification/regression problems
    if params["problem_type"] == "regression":
        overall_metrics = overall_stats(torch.Tensor([1]), torch.Tensor([1]), params)
    elif params["problem_type"] == "classification":
        # this is just used to generate the headers for the overall stats
        temp_tensor = torch.randint(0, params["model"]["num_classes"], (5,))
        overall_metrics = overall_stats(
            temp_tensor.to(dtype=torch.int32),
            temp_tensor.to(dtype=torch.int32),
            params,
        )

    metrics_log = params["metrics"].copy()
    if calculate_overall_metrics:
        for metric in overall_metrics:
            if metric not in metrics_log:
                metrics_log[metric] = 0

    # Setup a few loggers for tracking
    train_logger = Logger(
        logger_csv_filename=os.path.join(output_dir, "logs_training.csv"),
        metrics=metrics_log,
    )
    valid_logger = Logger(
        logger_csv_filename=os.path.join(output_dir, "logs_validation.csv"),
        metrics=metrics_log,
    )
    if testingDataDefined:
        test_logger = Logger(
            logger_csv_filename=os.path.join(output_dir, "logs_testing.csv"),
            metrics=metrics_log,
        )
    train_logger.write_header(mode="train")
    valid_logger.write_header(mode="valid")
    if testingDataDefined:
        test_logger.write_header(mode="test")

    if "medcam" in params:
        model = medcam.inject(
            model,
            output_dir=os.path.join(
                output_dir, "attention_maps", params["medcam"]["backend"]
            ),
            backend=params["medcam"]["backend"],
            layer=params["medcam"]["layer"],
            save_maps=False,
            return_attention=True,
            enabled=False,
        )
        params["medcam_enabled"] = False

    print("Using device:", device, flush=True)

    # Iterate for number of epochs
    for epoch in range(start_epoch, epochs):
        if params["track_memory_usage"]:
            file_to_write_mem = os.path.join(output_dir, "memory_usage.csv")
            if os.path.exists(file_to_write_mem):
                # append to previously generated file
                file_mem = open(file_to_write_mem, "a")
                outputToWrite_mem = ""
            else:
                # if file was absent, write header information
                file_mem = open(file_to_write_mem, "w")
                outputToWrite_mem = "Epoch,Memory_Total,Memory_Available,Memory_Percent_Free,Memory_Usage,"  # used to write output
                if params["device"] == "cuda":
                    outputToWrite_mem += "CUDA_active.all.peak,CUDA_active.all.current,CUDA_active.all.allocated"
                outputToWrite_mem += "\n"

            mem = psutil.virtual_memory()
            outputToWrite_mem += (
                str(epoch)
                + ","
                + str(mem[0])
                + ","
                + str(mem[1])
                + ","
                + str(mem[2])
                + ","
                + str(mem[3])
            )
            if params["device"] == "cuda":
                mem_cuda = torch.cuda.memory_stats()
                outputToWrite_mem += (
                    ","
                    + str(mem_cuda["active.all.peak"])
                    + ","
                    + str(mem_cuda["active.all.current"])
                    + ","
                    + str(mem_cuda["active.all.allocated"])
                )
            outputToWrite_mem += ",\n"
            file_mem.write(outputToWrite_mem)
            file_mem.close()

        # Printing times
        epoch_start_time = time.time()
        print("*" * 20)
        print("*" * 20)
        print("Starting Epoch : ", epoch)
        if params["verbose"]:
            print("Epoch start time : ", get_date_time())

        params["current_epoch"] = epoch

        epoch_train_loss, epoch_train_metric = train_network(
            model, train_dataloader, optimizer, params
        )
        epoch_valid_loss, epoch_valid_metric = validate_network(
            model, val_dataloader, scheduler, params, epoch, mode="validation"
        )

        patience += 1

        # Write the losses to a logger
        train_logger.write(epoch, epoch_train_loss, epoch_train_metric)
        valid_logger.write(epoch, epoch_valid_loss, epoch_valid_metric)

        if testingDataDefined:
            epoch_test_loss, epoch_test_metric = validate_network(
                model, test_dataloader, scheduler, params, epoch, mode="testing"
            )
            test_logger.write(epoch, epoch_test_loss, epoch_test_metric)

        if params["verbose"]:
            print("Epoch end time : ", get_date_time())
        epoch_end_time = time.time()
        print(
            "Time taken for epoch : ",
            (epoch_end_time - epoch_start_time) / 60,
            " mins",
            flush=True,
        )

        model_dict = get_model_dict(model, params["device_id"])

        # Start to check for loss
        if not (first_model_saved) or (epoch_valid_loss <= torch.tensor(best_loss)):
            best_loss = epoch_valid_loss
            best_train_idx = epoch
            patience = 0

            model.eval()

            save_model(
                {
                    "epoch": best_train_idx,
                    "model_state_dict": model_dict,
                    "optimizer_state_dict": optimizer.state_dict(),
                    "loss": best_loss,
                },
                model,
                params,
                model_paths["best"],
                onnx_export=False,
            )
            model.train()
            first_model_saved = True

        if params["model"]["save_at_every_epoch"]:
            save_model(
                {
                    "epoch": epoch,
                    "model_state_dict": model_dict,
                    "optimizer_state_dict": optimizer.state_dict(),
                    "loss": epoch_valid_loss,
                },
                model,
                params,
                os.path.join(
                    output_dir,
                    params["model"]["architecture"]
                    + "_epoch_"
                    + str(epoch)
                    + ".pth.tar",
                ),
                onnx_export=False,
            )
            model.train()

        # save the latest model
        if os.path.exists(model_paths["latest"]):
            os.remove(model_paths["latest"])
        save_model(
            {
                "epoch": epoch,
                "model_state_dict": model_dict,
                "optimizer_state_dict": optimizer.state_dict(),
                "loss": best_loss,
            },
            model,
            params,
            model_paths["latest"],
            onnx_export=False,
        )
        print("Latest model saved.")
        print("Current Best epoch: ", best_train_idx)

        if patience > params["patience"]:
            print(
                "Performance Metric has not improved for %d epochs, exiting training loop!"
                % (patience),
                flush=True,
            )
            break

    # End train time
    end_time = time.time()

    print(
        "Total time to finish Training : ",
        (end_time - start_time) / 60,
        " mins",
        flush=True,
    )

    # once the training is done, optimize the best model
    if os.path.exists(model_paths["best"]):
        optimize_and_save_model(model, params, model_paths["best"], onnx_export=True)


if __name__ == "__main__":
    import argparse, pickle

    torch.multiprocessing.freeze_support()
    # parse the cli arguments here
    parser = argparse.ArgumentParser(description="Training Loop of GANDLF")
    parser.add_argument(
        "-train_loader_pickle", type=str, help="Train loader pickle", required=True
    )
    parser.add_argument(
        "-val_loader_pickle", type=str, help="Validation loader pickle", required=True
    )
    parser.add_argument(
        "-testing_loader_pickle",
        type=str,
        help="Testing loader pickle",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-parameter_pickle", type=str, help="Parameters pickle", required=True
    )
    parser.add_argument("-outputDir", type=str, help="Output directory", required=True)
    parser.add_argument("-device", type=str, help="Device to train on", required=True)

    args = parser.parse_args()

    # # write parameters to pickle - this should not change for the different folds, so keeping is independent
    parameters = pickle.load(open(args.parameter_pickle, "rb"))
    trainingDataFromPickle = pd.read_pickle(args.train_loader_pickle)
    validationDataFromPickle = pd.read_pickle(args.val_loader_pickle)
    testingData_str = args.testing_loader_pickle
    testingDataFromPickle = pd.read_pickle(testingData_str) if testingData_str else None

    training_loop(
        training_data=trainingDataFromPickle,
        validation_data=validationDataFromPickle,
        output_dir=args.outputDir,
        device=args.device,
        params=parameters,
        testing_data=testingDataFromPickle,
    )



from typing import Optional, Tuple
import torch
import psutil
from .loss_and_metric import get_loss_and_metrics


def step(
    model: torch.nn.Module,
    image: torch.Tensor,
    label: torch.Tensor,
    params: dict,
    train: Optional[bool] = True,
) -> Tuple[float, dict, torch.Tensor, torch.Tensor]:
    """
    This function performs a single step of training or validation.

    Args:
        model (torch.nn.Module): The model to process the input image with, it should support appropriate dimensions.
        image (torch.Tensor): The input image stack according to requirements.
        label (torch.Tensor): The input label for the corresponding image tensor.
        params (dict): The parameters dictionary.
        train (Optional[bool], optional): Whether the step is for training or validation. Defaults to True.

    Returns:
        Tuple[float, dict, torch.Tensor, torch.Tensor]: The loss, metrics, output, and attention map.
    """
    if params["verbose"]:
        if torch.cuda.is_available():
            print(torch.cuda.memory_summary())
        print(
            "|===========================================================================|"
        )
        print(
            "|                              CPU Utilization                              |"
        )
        print("Load_Percent   :", psutil.cpu_percent(interval=None))
        print("MemUtil_Percent:", psutil.virtual_memory()[2])
        print(
            "|===========================================================================|"
        )

    # for the weird cases where mask is read as an RGB image, ensure only the first channel is used
    if label is not None:
        if params["problem_type"] == "segmentation":
            if label.shape[1] == 3:
                label = label[:, 0, ...].unsqueeze(1)
                # this warning should only come up once
                if params["print_rgb_label_warning"]:
                    print(
                        "WARNING: The label image is an RGB image, only the first channel will be used.",
                        flush=True,
                    )
                    params["print_rgb_label_warning"] = False

            if params["model"]["dimension"] == 2:
                label = torch.squeeze(label, -1)

    if params["model"]["dimension"] == 2:
        image = torch.squeeze(image, -1)
        if "value_keys" in params:
            if label is not None:
                if len(label.shape) > 1:
                    label = torch.squeeze(label, -1)

    if not (train) and params["model"]["type"].lower() == "openvino":
        output = torch.from_numpy(
            model(inputs={params["model"]["IO"][0][0]: image.cpu().numpy()})[
                params["model"]["IO"][1][0]
            ]
        )
        output = output.to(params["device"])
    else:
        if params["model"]["amp"]:
            with torch.cuda.amp.autocast():
                output = model(image)
        else:
            output = model(image)

    attention_map = None
    if "medcam_enabled" in params and params["medcam_enabled"]:
        output, attention_map = output

    # one-hot encoding of 'label' will probably be needed for segmentation
    if label is not None:
        loss, metric_output = get_loss_and_metrics(image, label, output, params)
    else:
        loss, metric_output = None, None

    if len(output) > 1:
        output = output[0]

    if params["model"]["dimension"] == 2:
        output = torch.unsqueeze(output, -1)
        if "medcam_enabled" in params and params["medcam_enabled"]:
            attention_map = torch.unsqueeze(attention_map, -1)

    return loss, metric_output, output, attention_map
